package cz.zcu.fav.kiv.jsimcasestudies.watersystem.application;

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsimcasestudies.watersystem.*;


/**
 *  Model of a thread that is responsible for controlling the pump of a water
 *  source.
 *
 *  !!! This source code uses JSim directly (passivate() in waitChange()) 
 *      and it should be changed (it should use the JiJ package). 
 *       
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @author St. RACEK
 *  @version 1.1 from 2004-03-18
 */

class SourceControlThread extends JSimProcess
{	       
        
        /**
	 *  The source that this thread manages.
	 */
	private final SourceMonitor source;

        /**
	 *  The station that manages this thread.
	 */
	private final StationMonitor station;
       

        /**
	 *  The object that implements interrupt entry
	 */
	private final InterruptEntry iE;


        /**
	 *  The int-coded state of control thread - values:
         *     0  ... (BlOff) water below low level, pump off
         *     1  ... (BlOn)  water below low level, pump on   (transient state)
         *     2  ... (LlOff) water between levels, pump off
         *     3  ... (LlOn)  water between levels, pump on
         *     4  ... (HlOff) water above high level, pump off
         *     5  ... (HlOn)  Water above high level, pump on
	 */


	/* --------------------------------------------------------------------- */

	/**
	 *  Constructs a new thread of the type SourceControlThread
	 *
	 *  @param simulation	The overall simulation.
         *  @param station	The station that manages this process.
	 *  @param source	The source to be managed by this process.
         *  @param iI           The interupt interface to bind the service routine 	
	 */
	protected SourceControlThread(JSimSimulation simulation, SourceMonitor source,
          StationMonitor station,  InterruptEntry iE)
          throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException,
          JSimTooManyProcessesException {
	    super("Water Source Process #" + String.valueOf(source.getNumber()), simulation);
	    this.source = source;
            this.station = station;
            this.iE = iE;                         
            source.pumpSwitchOff();
	} // constructor


        /**
         *  Creates integer code of the source state
         *     - see the coding of states given above
	 */

        private int getControlState () {
          SourceState s = source.getSourceState();
          if (!s.levelLow) {
            if (s.pumpIsOn)
              return 1;
            else
              return 0;
          }
          if (s.levelHigh) {
            if (s.pumpIsOn)
              return 5;
            else
              return 4;
          }
          if (s.pumpIsOn)
            return 3;
          else
            return 2;
        } // getControlState


        /**
         *  Waits for a sensor change
	 */
        private void waitChange () throws JSimSecurityException{          
          iE.waitInterrupt(this);         
          passivate();   
        } 

        /**
         *  Endless activity that periodically tests the source state and 
         *  switch on/off the pump according to water levels in the source and in the tank.
	 */

	protected void life() {
          StationState state;
	  try {
	    
            while (true) {
              switch (getControlState()) {
                case 0:   // water below source low level, pump off
                  break;
                case 1:   // water below source low level, pump on
                  source.pumpSwitchOff();
                  station.releaseActivity();
                  break;
                case 2:   // water between source levels, pump off                  
                  break;
                case 3:   // water between source levels, pump on
                  state = station.getStationState();
                  if (state.levelHigh) {
                    source.pumpSwitchOff();
                    station.releaseActivity();
                  }
                  break;
                case 4:   // water above source high level, pump off
                  if (station.requestActivity())
                    source.pumpSwitchOn();
                  break;
                case 5:   // Water above source high level, pump on
                  state = station.getStationState();
                  if (state.levelHigh) {
                    source.pumpSwitchOff();
                    station.releaseActivity();
                  }
                  break;
                default: break;
              }                            

              // active waiting - without interrupts
              // hold (1.0);

              // passive waiting for an interrupt
              waitChange();
             

            } // while true
          } // try

          catch (JSimException e){
	    e.printComment(System.err);
	    e.printStackTrace();
	  } // catch
        }  // life

} // class SourceControlThread

