package cz.zcu.fav.kiv.jsimcasestudies.watersystem.application;

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsimcasestudies.watersystem.*;
import cz.zcu.fav.kiv.jsimcasestudies.watersystem.model.*;
import java.io.*;


/*
 *  Model based implementation of Water station Control application
 *    - main program.
 *  It has to be started:
 *  WaterStation maxTime runMode
 *     where:
 *        maxTime is the overall time of simulation in seconds
 *        runMode : value 0 - time, check of invariants, current output flow and
 *                            volume are printed in every simulation step,
 *                            check of invariants doesn't influence flow of
 *                            computation     
 *                  value 1 - like 0, but state of model and control part is added,
 *                            invalid invariant passes computation to a step-by-step
 *                            mode (similar like runMode = 2)
 *                  value 2 - like 1, but simulation step is triggered by a key 'q'
 *                            means "to finish"
 *
 *
 *  !!! This source code should be rewritten substantialy when converted to the 
 *     production version.  
 *
 *
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @author St. Racek
 *  @version 1.1 from 2004-03-20
 *
 */





public class WaterStation  {

 /**
  *  Number of water sources in the system.
  */
  private static final int N           = 3;


  /**
   *  Maximum number of pumps that is allowed to run simultaneously.
   */
  private static final int K           = 2;


  /**
   *  Base period of control loops.
   */
  private static final double PERIOD   = 1.0; //seconds

  /**
  * A method that read a char value
  */

public static char readChar () {
  try {
    return (char) System.in.read();
  }
  catch (IOException e) {
  return ('\uFFFF');
  }
}

  /**
  * Main method of the WaterStation class
  */
public static void main(String[] args) {

  JSimSimulation simulation = null;
  CommonControlInterface controlInterface = null;
  WaterStationControl controlProgram = null;
  //ModelInterface modelProgram = null;

  try
  {
    /* First, we create a simulation object. */
    simulation = new JSimSimulation("Water Station Simulation");

    /* Second, we create a simulation world of environment -
       it includes activation of its processes */
    controlInterface = new ModelInterface(simulation, N, PERIOD);

    /* For production version of control program another object
     *   of interface should be created, for example:
     * controlInterface = new NativeInterface (baseHwAddress, N, PERIOD);
     */

    /* Third, we create a simulation world of the control application -
       it includes activation of its processes */
    controlProgram = new WaterStationControl (simulation, controlInterface, N, K, PERIOD);

    /* setting length of simulation in seconds of model time */
    double maxTime = (double) (Integer.valueOf(args[0]).intValue());
    if (maxTime < 0.0) maxTime = 100.0;
    double time = 0.0;

    /* setting the simulation run mode ) */
    int runMode = Integer.valueOf(args[1]).intValue();
    if (runMode < 0 || runMode > 2) {
      System.out.println ("Bad value of runMode!!!");
      return;
    }

    /* Now we run the simulation. */
    while ((simulation.step() == true)&&(time<=maxTime)) {

      int test = ((ModelInterface)controlInterface).checkInvariants(K);       

      time = simulation.getCurrentTime();
      System.out.println ("Time: " +   time + "  invariants: " + test
        + "  volume: " + ((ModelInterface) controlInterface).readVolume()
        + "  output flow: " + ((ModelInterface) controlInterface).readOutputFlow());

      int runModeTmp;
      if (test > 0 && runMode > 0)
        runModeTmp = 2;
      else 
        runModeTmp = runMode;

      switch (runModeTmp) {
         case 0: break;
         case 1: 
                 controlProgram.displayState();                
                 ((ModelInterface) controlInterface).displayState();
                 break;        
         case 2: 
                 controlProgram.displayState();
                 ((ModelInterface) controlInterface).displayState();
                 char c = readChar();                 
                 // Pressed 'q' key you quit from the application
                   if (c == 'q') return;

                 break;
         default:break;
      }
    }
  } // try

  /*   All three possible exceptions are caught here in one handler
   */
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch

  /*
   *  It is a good practice to put the shutdown in the finally block. Then you can be sure
   *  that that the JVM will not "hang". (This wouldn't occur here.)
   */
  finally
  {
   /*
    *  Every simulation must be shut down at the end of the program, even if an error occurs.
    */
   simulation.shutdown();
  }
 } // main
} // WaterStation


