package cz.zcu.fav.kiv.jsimcasestudies.watersystem.application;

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsimcasestudies.watersystem.*;

/**
 *  An encapsulation of the whole control program part of the simulation.
 *  It contains data models (monitors) of water sources as well as the data model
 *  (monitor) of the water station.
 *  It uses an object of the type CommonControlInterface in order to access to
 *  the model of its controlled environment.
 *  It also incorporates threads (processes) responsible for controlling pumps.
 *
 *  !!! This source code needs to be edited when passing it to the 
 *      production version.  
 * 
 *  !!! Functions startControl() and stopControl() should be added.  
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @author St. Racek
 *  @version 1.1 from 2004-03-18
 */

public class WaterStationControl
{
	/**
	 *  Base period of control loops.
	 */
	private double basePeriod;

	/**
	 *  Number of water sources in the system.
	 */
	private int numberOfSources;

	/**
	 *  Maximum number of pumps that is allowed to run simultaneously.
	 */
	private int maxNumberOfPumpsOn ;

        /**
	 *  Sources objects.
	 */
	private SourceMonitor[] source;

	/**
	 *  The main station object.
	 */
	private StationMonitor station;

	/**
	 *  Process-thread controlling the water station
	 */
	//private StationCoordinatingThread coordinator;

	/**
	 *  Processes-threads controlling the sources
	 */
	private SourceControlThread[] sourceControl;

	/**
	 *  The simulation that this model is a part of.
	 */
	private JSimSimulation simulation;

        /**
	 *  An instance of CommonControlInterface.
	 */
	private CommonControlInterface controlInterface;

        /**
	 *  An instance of InterruptController Interface, it represents HW interrupt controller
	 */
	private InterruptControllerInterface iI;

        /**
	 *  An instance of InterruptEntry, it encapsulates interrupt service routine.
	 */
	private InterruptEntry iE;


	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Constructs a new control program with specified number of water sources.
	 *  It creates all necessary threads and monitors.
	 *
	 *  @param simulation The overall simulation.
         *  @param controlInterface The object of control interface
	 *  @param numberOfSources Number of water sources in the simulation.
         *  @param numberOfPumpsOn Maximum number of pumps to run simultaneously
         *  @param basePeriod Base period of active control loops
	 */
	public WaterStationControl(JSimSimulation simulation, CommonControlInterface controlInterface,
          int numberOfSources, int maxNumberOfPumpsOn, double basePeriod)
        throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException,
        JSimTooManyProcessesException, JSimTooManyHeadsException, JSimSecurityException
	{
		this.simulation = simulation;
                this.controlInterface = controlInterface;
                this.numberOfSources = numberOfSources;
                this.maxNumberOfPumpsOn = maxNumberOfPumpsOn;
                this.basePeriod = basePeriod;

                // Station monitor creation an setup
                station = new StationMonitor(controlInterface, maxNumberOfPumpsOn);

                // Source monitors creation an setup
		source = new SourceMonitor[numberOfSources];
		for (int i = 0; i < source.length; i++)
		  source[i] = new SourceMonitor(controlInterface, i);

                // Reading the adress of interrupt controller device
                iI = controlInterface.getInterruptControllerAddress();
                
                // Creating an entry of interrupt (here it is one common interrupt
                //   for all changes in environment). Interrupt entry
                //   contains interrupt service routine as well. 
                iE = new InterruptEntry (simulation);

                // Binding HW interrupts to its service routines 
                //   (here is only one common service routine)
                iI.bindInterrupt(0, iE);    // level 0 - timer interrupt
                iI.bindInterrupt(1, iE);    // level 1 - tank sensors (change false->true)
                iI.bindInterrupt(2, iE);    // level 2 - sources sensors (change false->true)    

		// Source Control Processes setup and activation
		sourceControl = new SourceControlThread[numberOfSources];
		for (int i = 0; i < sourceControl.length; i++)
		{
			sourceControl[i] = new SourceControlThread(simulation, source[i],
                          station, iE);
			sourceControl[i].activate(simulation.getCurrentTime() + 0.1 * i);
		} // for

		//coordinator = new StationCoordinatingThread (simulation, station, basePeriod);
		//coordinator.activate(simulation.getCurrentTime());

	} // constructor

	/**
	* Displays the state-like data on a console
	*
	*/
        public void displayState() {  // prints the state-like data on a console
          System.out.print ("Sources: ");
          for (int i = 0; i < source.length; i++)
            (source[i].getSourceState()).displayState();
          System.out.print ("Station: ");
          (station.getStationState()).displayState();
          System.out.println();
        }

        // TBD methods:

        // public void stopControl();   // stops the run of control program, the controlled
              // object should be after it within a safe state
        // similarly startControl(); - when started after the stop() call

} // class WaterStationControl

