package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsimcasestudies.watersystem.*;



/**
 *  An encapsulation of the whole environment model part of the simulation program.
 *  It provides access to the model of hardware data values that will be read/set
 *  by the tested control program instead of real hardware registers/memory.
 *  It also incorporates processes responsible for dynamic modeling 
 *  the controlled object and hardware sensors that lies between the 
 *  controlling computer and controlled object.
 *
 *  !!! It implements the abstract CommonControlInterface used by the control program.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.2 from 2004-03-19
 */


public class ModelInterface implements CommonControlInterface
{
	/**
	 *  Base period of data updates.
	 */
	private double basePeriod;

	/**
	 *  Number of water sources in the system.
	 */
	private int numberOfSources;

	/**
	 *  Sensor data of all sources.
	 */
	private SourceHardwareData[] sourceData;

	/**
	 *  Sensor data of the main tank.
	 */
	private StationHardwareData stationData;

        /**
	 *  Interrupt controller object.
	 */
	private InterruptController ic;

	/**
	 *  Source objects.
	 */
	private Source[] sources;

	/**
	 *  The main station object.
	 */
	private Station station;

	/**
	 *  Processes simulating water volume changes in the sources.
	 */
	private SourceProcess[] sourceProcesses;

	/**
	 *  Process simulating water volume changes in the main tank.
	 */
	private StationProcess stationProcess;

        /**
	 *  Process simulating HW timer device
	 */
	private TimerProcess timer;

	/**
	 *  Process simulating water consumption and water volume changes in the main tank.
	 */
	private WaterConsumptionProcess consumptionProcess;

	/**
	 *  The simulation that this model is a part of.
	 */
	private JSimSimulation simulation;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Constructs a new interface of a model with specified number of water sources.
	 *  Also creates all necessary processes.
	 *
	 *  @param simulation The overall simulation.
	 *  @param numberOfSources Number of water sources in the simulation.
	 */
	public ModelInterface(JSimSimulation simulation, int numberOfSources, double basePeriod) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException, JSimSecurityException
	{
		this.numberOfSources = numberOfSources;
		this.simulation = simulation;
		this.basePeriod = basePeriod;

		double currentSimulationTime = simulation.getCurrentTime();

		// Hardware data setup
		sourceData = new SourceHardwareData[numberOfSources];
		for (int i = 0; i < sourceData.length; i++)
			sourceData[i] = new SourceHardwareData();

		stationData = new StationHardwareData();
                
                // Interrupt controller creation, 4 intterupts, initially enabled
                ic = new InterruptController(4, false);
		
                // HW timer device creation creation
                timer = new TimerProcess(simulation, ic, basePeriod);
                timer.activate(currentSimulationTime);


                // Source objects setup and creation
		sources = new Source[numberOfSources];
		for (int i = 0; i < sources.length; i++)
			sources[i] = new Source(i);

		station = new Station(sourceData);

		// Processes setup and activation
		sourceProcesses = new SourceProcess[numberOfSources];
		for (int i = 0; i < sources.length; i++)
		{
			sourceProcesses[i] = new SourceProcess(simulation, sources[i], sourceData[i],                                                                ic, basePeriod);
			sourceProcesses[i].activate(currentSimulationTime + 0.2 * i);
		} // for

		stationProcess = new StationProcess(simulation, stationData, station,
                                                    ic, basePeriod);
		stationProcess.activate(currentSimulationTime);

		consumptionProcess = new WaterConsumptionProcess(simulation, stationData,
                                                                 station, basePeriod);
		consumptionProcess.activate(currentSimulationTime);
	} // constructor

	/**
	 *  Reads the Source Level Low flag of a water source.
	 *
	 *  @param sourceNo Number of the source whose flag has to be read.
	 *
	 *  @return The value of the Source Level Low flag of the specified source.
	 */
	public boolean readSourceLevelLow(int sourceNo)
	{
		return sourceData[sourceNo].getSourceLevelLow();
	}

	/**
	 *  Reads the Source Level High flag of a water source.
	 *
	 *  @param sourceNo Number of the source whose flag has to be read.
	 *
	 *  @return The value of the Source Level High flag of the specified source.
	 */
	public boolean readSourceLevelHigh(int sourceNo)
	{
		return sourceData[sourceNo].getSourceLevelHigh();
	}

	/**
	 *  Reads the Pump IS On flag of a water source.
	 *
	 *  @param sourceNo Number of the source whose flag has to be read.
	 *
	 *  @return The value of the Pump Is On flag of the specified source.
	 */
	public boolean readPump(int sourceNo)
	{
		return sourceData[sourceNo].getPumpIsOn();
	}

	/**
	 *  Sets the Pump Is On flag of a water source.
	 *
	 *  @param sourceNo Number of the source whose flag has to be set.
	 *
	 *  @return The old value of the Pump Is On flag of the specified source.
	 */
	public boolean setPump(int sourceNo, boolean newState)
	{
		boolean oldState;

		oldState = sourceData[sourceNo].getPumpIsOn();
		sourceData[sourceNo].setPumpIsOn(newState);

		return oldState;
	}

	/**
	 *  Reads the Source Level Low flag of the main water tank.
	 *
	 *  @return The value of the Source Level Low flag of the main water tank.
	 */
	public boolean readWaterStationLevelLow()
	{
		return stationData.getStationLevelLow();
	}

	/**
	 *  Reads the Station Level High flag of the main water tank.
	 *
	 *  @return The value of the Station Level High flag of the main water tank.
	 *
	 */
	public boolean readWaterStationLevelHigh()
	{
		return stationData.getStationLevelHigh();
	}

	/**
	 *  Reads the Output Flow register of the main water tank.
	 *
	 *  @return The value of the Output Flow register of the main water tank.
	 */
	public int readOutputFlow()
	{
		return stationData.getOutputFlow();
	}
        
        /**
	 *  Reads the Volume of the main water tank.
	 *
	 *  @return The value of volume of the main water tank.
	 */
	public double readVolume()
	{
		return station.getVolume();
	}
 
 	/**
	 *  Returns the reference to the interrupt controller object.
	 *
	 *  @return The reference to the interrupt controller object.
	 */
	public InterruptControllerInterface getInterruptControllerAddress() 
        {
          return ic;
        }             
       
 
       /**
	 *  Checks state invariants of the modeled system 
	 *
         *  @param k - limiting number of pumps to be switched on simultaneously 
	 *  @return zero if (all) invariants are valid else return an integer value 
         *    - number of invariant to be invalid
	 */
	public int checkInvariants(int k)
	{
	  int nPumpsOn = station.howManyPumpsOn();
 
         /* invariant no.1 - number of running pumps should not exceed k value */
          if (nPumpsOn > k) return 1;     

         /* invariant no.2 - if the tank volume is at high level, no pump should run */
          if (readWaterStationLevelHigh() && nPumpsOn>0) return 2;

         /* invariant no.3 - if any source is below low level, its pump must
          be switched off */ 
          for (int i = 0; i<numberOfSources; i++)
            if(!readSourceLevelLow(i) &&  readPump(i)) return 3;
 
         /* how many sources is at high level */
          int nSourcesHigh = 0; 
          for (int i = 0; i<numberOfSources; i++)
           if(readSourceLevelHigh(i)) nSourcesHigh++;

         /* invariant no.4 - if the tank volume is below low level and number of sources
            at high level is higher than the limit k, then the number of running pumps
            should equal k */           
          if(!readWaterStationLevelLow() && nSourcesHigh >= k && !(nPumpsOn == k))
            return 4;

         /* all invariants are valid */
          return 0;

	}
               

	/**
	*  Displays the state of the model at a console. 
	*/

	public void displayState() {  // prints the state-like data on a console
		System.out.print ("Sources: ");
		for (int i = 0; i < sources.length; i++)
			sourceData[i].displayState();
		System.out.print ("Station: ");
		stationData.displayState();
		System.out.print(" ? ? ");
                System.out.print(station.getVolume() + "  " + readOutputFlow());
		System.out.println();
        }


	/**
	 *  This method allows the model part to be simulated alone without the tested application.
	 *  This is important in order to verify the model itself.
	 *
	 *  @param args Command line arguments.
	 */
	public static void main(String[] args)
	{
		JSimSimulation simulation = null;
		ModelInterface modelInterface = null;

		try
		{
			simulation = new JSimSimulation("Model Simulation", JSimSimulation.MODE_GUI_INTERACTIVE);
			modelInterface = new ModelInterface(simulation, (int) 5, 0.5);
			simulation.runGUI();
		} // try
		catch (JSimException e1)
		{
			e1.printComment(System.err);
			e1.printStackTrace();
		} // catch
		finally
		{
			simulation.shutdown();
		} // finally
	} // main


} // class ModelInterface

