package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

//package cz.zcu.fav.kiv.jsimcasestudies.watersystem.application;

/**

 *  A source maintains the volume of water it holds inside. It also knows its minimum and maximum volume,
 *  as well as the maximum natural volume of water it can absorb.
 *  It is the data model of water source.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.1 from 2004-03-24
 */
class Source
{
	/**
	 *  The lowest possible volume of water in the source, in liters. If the actual volume is below this
	 *  limit, the hardware sensor will indicate low level.
	 */
	protected static final double LOW_VOLUME_LIMIT = 40.0;

	/**
	 *  The highest possible volume of water in the source, in liters. If the actual volume is above
	 *  this limit, the hardware sensor will indicate high level.
	 */
	protected static final double HIGH_VOLUME_LIMIT = 300.0;

	/**
	 *  The maximal volume of water (in liters) that the source can ever absorb.
	 *  The Source Level High sensor is placed below this limit so the pump switches on
	 *  before the source is actually completely full. This limit is an environmental constraint.
	 */
	protected static final double NATURAL_MAX_VOLUME_LIMIT = 500.0;

	/**
	 *  The quantity of water (in liters per second) pumped out from the source in one second,
	 *  if the pump is running.
	 */
	protected static final double PUMP_POWER_PER_SECOND = 10.0;

	/**
	 *  The maximum quantity of water that can come to the source from the outer environment,
	 *  in liters per second.
	 */
	protected static final double MAX_VOLUME_DELTA_PER_SECOND = 4.0;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  The number of the source.
	 */
	private final int sourceNumber;

	/**
	 *  Volume of water currently present in the source.
	 */
	private double currentVolume;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Creates a new source object and sets its volume to the exact middle of possible values.
	 *
	 *  @param sourceNumber	Number of the source.
	 */
	protected Source(int sourceNumber)
	{
		this.sourceNumber = sourceNumber;
		currentVolume = LOW_VOLUME_LIMIT + (HIGH_VOLUME_LIMIT - LOW_VOLUME_LIMIT) / 2;
	} // constructor

	/**
	 *  Returns the source's number.
	 *
	 *  @return The source's number.
	 */
	protected int getNumber()
	{
		return sourceNumber;
	} // getNumber

	/**
	 *  Increases the volume of water in the source.
	 *
	 *  @param delta The amount of water that is added to the source.
	 */
	protected void increaseVolume(double delta)
	{
		currentVolume += delta;
		if (currentVolume > NATURAL_MAX_VOLUME_LIMIT)
			currentVolume = NATURAL_MAX_VOLUME_LIMIT;
	} // increaseVolume

	/**
	 *  Decreases the volume of water in the source.
	 *
	 *  @param delta The amount of water that is taken away from the source.
	 */
	protected void decreaseVolume(double delta)
	{
		currentVolume -= delta;
		if (currentVolume < 0)
			currentVolume = 0;
	} // decreaseVolume

	/**
	 *  Reports the current volume of water in the source.
	 *
	 *  @return The current volume of water in the source.
	 */
	protected double getVolume()
	{
		return currentVolume;
	} // getVolume


} // class Source


