package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

import cz.zcu.fav.kiv.jsim.*;

/**
 *  The dynamic model of water source (i.e. its state - water level) changes.
 *  This process is responsible for periodic adjustments of model values concerning a particular water source.
 *  This involves switching on/off sensors of low/high level of water in the source.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.3 from 2004-03-24
 */
class SourceProcess extends JSimProcess
{
	/**
	 *  Multiplier of the model's base period. The product of this multiplier and the model's base
	 *  period will be used as a parameter to the hold() method.
	 */
	private static final double BASE_PERIOD_MULTIPLIER = 1.0;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  The model's base period of updates.
	 */
	private final double basePeriod;

	/**
	 *  The source that this process manages.
	 */
	private final Source source;

	/**
	 *  Information produced by hardware sensors.
	 */
	private final SourceHardwareData hw;

        /**
	 *  Interrupt controller object.
	 */
	private InterruptController ic;


	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Constructs a new process simulating environment changes and hardware activity that concerns
	 *  the given source.
	 *
	 *  @param simulation	The overall simulation.
	 *  @param source	The source to be managed by this process.
	 *  @param hw		Hardware data periodically changed by this source's sensors.
         *  @param ic           Interrupt object (flag) sets with any sensor change
	 *  @param basePeriod	The model's base period.
	 */
	protected SourceProcess(JSimSimulation simulation, Source source, SourceHardwareData hw, 
              InterruptController ic, double basePeriod) throws JSimSimulationAlreadyTerminatedException,               JSimInvalidParametersException, JSimTooManyProcessesException
	{
		super("Water Source Process #" + String.valueOf(source.getNumber()), simulation);
		this.source = source;
		this.hw = hw;
                this.ic = ic;
		this.basePeriod = basePeriod;
	} // constructor

	/**
	 *  Simulates water level changes in the source and sensor setting. Can set the sourceLevelLow
	 *  and sourceLevelHigh flags.
	 */
	protected void life()
	{
		try
		{
			double holdTime = basePeriod * BASE_PERIOD_MULTIPLIER;
                        boolean sensorOldValue, sensorNewValue; 

			while (true)
			{
				// message(getProcessName() + ": Holding for " + holdTime);
				hold(holdTime);

				source.increaseVolume(JSimSystem.uniform(0.0, Source.MAX_VOLUME_DELTA_PER_SECOND) * holdTime);
				if (hw.getPumpIsOn())
					source.decreaseVolume(Source.PUMP_POWER_PER_SECOND * holdTime);
				// message(getProcessName() + ": Volume in the source = " + source.getVolume() + " liters.");
				/* if (hw.getPumpIsOn())
					message(getProcessName() + ": Pump is running.");
				else
					message(getProcessName() + ": Pump is not running.");
                                */

                                // setting the sensors

				sensorOldValue = hw.getSourceLevelLow();
                                hw.setSourceLevelLow(source.getVolume() >= Source.LOW_VOLUME_LIMIT);
                                sensorNewValue = hw.getSourceLevelLow();
                                if (sensorOldValue != sensorNewValue) ic.setInterrupt(2);

                                sensorOldValue = hw.getSourceLevelHigh();
				hw.setSourceLevelHigh(source.getVolume() >= Source.HIGH_VOLUME_LIMIT);
                                sensorNewValue = hw.getSourceLevelHigh();
                                if (sensorOldValue != sensorNewValue) ic.setInterrupt(2);

				// message(getProcessName() + ": SLL=" + hw.getSourceLevelLow() + " SLH=" + hw.getSourceLevelHigh());
			} // while true
		} // try
		catch (JSimSecurityException e1)
		{
			e1.printComment(System.err);
			e1.printStackTrace();
		} // catch
	} // life

} // class SourceProcess

