package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

/**
 *  A station object maintains the volume of water in the main tank.
 *  It is the data model of the main tank.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.1 from 2004-03-19
 */

class Station
{
	/**
	 *  The lowest possible volume of water in the source, in liters. If the actual volume is below this
	 *  limit, the hardware sensor will indicate low level.
	 */
	protected static final double LOW_VOLUME_LIMIT = 1000.0;

	/**
	 *  The highest possible volume of water in the source, in liters. If the actual volume is above
	 *  this limit, the hardware sensor will indicate high level.
	 */
	protected static final double HIGH_VOLUME_LIMIT = 3000.0;

	/**
	 *  The maximum output flow in liters per second. This constant is determined by the size
	 *  of the output pipe.
	 */
	protected static final double MAX_OUTPUT_FLOW = 12.0;

	/**
	 *  The maximum change of output flow per one second. The actual change can be either positive
	 *  or negative.
	 */
	protected static final double MAX_OUTPUT_FLOW_CHANGE_PER_SECOND = 1.0;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Volume of water currently present in the main tank.
	 */
	private double currentVolume;

	/* ------------------------------------------------------------------------------------------------ */

        /**
	 *  Sources' hardware info. This is necessary for counting the current input flow from all sources
	 *  whose pumps are currently switched on. Will be read only, not modified.
	 */
	private SourceHardwareData[] sources;

	/**
	 *  Creates a new station object and sets its volume to the exact middle of possible values.
	 */
	protected Station(SourceHardwareData[] sources)
	{
	    this.sources = sources;	
            currentVolume = LOW_VOLUME_LIMIT + (HIGH_VOLUME_LIMIT - LOW_VOLUME_LIMIT) / 2;
	} // constructor

	/**
	 *  Increases the volume of water in the tank. The upper limit is not specified.
	 *
	 *  @param delta The amount of water that is added to the tank.
	 */
	protected void increaseVolume(double delta)
	{
		currentVolume += delta;
	} // increaseVolume

	/**
	 *  Decreases the volume of water in the tank.
	 *
	 *  @param delta The amount of water that is taken away from the tank.
	 */
	protected void decreaseVolume(double delta)
	{
		currentVolume -= delta;
		if (currentVolume < 0)
			currentVolume = 0;
	} // decreaseVolume

	/**
	 *  Reports the current volume of water in the tank.
	 *
	 *  @return The current volume of water in the tank.
	 */
	protected double getVolume()
	{
		return currentVolume;
	} // getVolume


	/**
	 *  Returns the number of pumps currently switched on. Uses hardware data of all water sources
	 *  of the model.
	 *
	 *  @return The number of pumps currently switched on.
	 */
	protected int howManyPumpsOn()
	{
		if ((sources == null) || (sources.length < 1))
			return 0;
		else
		{
			int count = 0;
			for (int i = 0; i < sources.length; i++)
				if ((sources[i] != null) && (sources[i].getPumpIsOn()))
					count++;

			return count;
		} // else
	} // howManyPumpsOn


} // class Station


