package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

import cz.zcu.fav.kiv.jsim.*;

/**
 *  The dynamic model of water station (i.e. its state - main tank water level) changes.
 *  This process is responsible for periodic adjustments of model values concerning the main water tank.
 *  This involves switching on/off sensors of low/high level of water in the tank. It is assumed that just
 *  one process of this class will be created in the whole model.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.3 from 2004-03-24
 */

class StationProcess extends JSimProcess
{
	/**
	 *  Multiplier of the model's base period. The product of this multiplier and the model's base
	 *  period will be used as a parameter to the hold() method.
	 */
	private static final double BASE_PERIOD_MULTIPLIER = 1.0;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  The model's base period of updates.
	 */
	private final double basePeriod;

	/**
	 *  Information produced by hardware sensors.
	 */
	private StationHardwareData hw;

        /**
	 *  Interrupt controller object.
	 */
	private InterruptController ic;
	

        /**
	 *  The station that this process manages.
	 */
	private final Station station;

	

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Constructs a new process simulating water changes in the main tank. Hardware sensors stationLevelLow
	 *  and stationLevelHigh are set by this process.
	 *
	 *  @param simulation	The overall simulation.
	 *  @param hw		Hardware data periodically changed by the tank's sensors.
	 *  @param station	The station that this process will manage.
         *  @param ic           Interrupt object (flag) sets with any sensor change	 
	 *  @param basePeriod	The model's base period.
	 */
	protected StationProcess(JSimSimulation simulation, StationHardwareData hw, Station station, 
           InterruptController ic, double basePeriod) throws JSimSimulationAlreadyTerminatedException,                       JSimInvalidParametersException, JSimTooManyProcessesException
	{
		super("Water Station Process", simulation);
		this.hw = hw;
		this.station = station;
                this.ic = ic;		
		this.basePeriod = basePeriod;
	} // constructor

	/**
	 *  Simulates water level changes in the station tank and sensor setting. Can set the sourceLevelLow
	 *  and sourceLevelHigh flags.
	 */
	protected void life()
	{
		try
		{
			int numberOfPumpsOn;
			double holdTime = basePeriod * BASE_PERIOD_MULTIPLIER;
                        boolean sensorOldValue, sensorNewValue; 

			while (true)
			{
				// message(getProcessName() + ": Holding for " + holdTime);
				hold(holdTime);

				numberOfPumpsOn = station.howManyPumpsOn();
				station.increaseVolume(Source.PUMP_POWER_PER_SECOND * numberOfPumpsOn * holdTime);
				// message(getProcessName() + ": Volume in the main tank = " + station.getVolume() + " liters.");

				// setting the sensors

                                sensorOldValue = hw.getStationLevelLow();
                                hw.setStationLevelLow(station.getVolume() >= Station.LOW_VOLUME_LIMIT);
                                sensorNewValue = hw.getStationLevelLow();
                                if (sensorOldValue != sensorNewValue) ic.setInterrupt(1);

                                sensorOldValue = hw.getStationLevelHigh();
				hw.setStationLevelHigh(station.getVolume() >= Station.HIGH_VOLUME_LIMIT);
                                sensorNewValue = hw.getStationLevelHigh();                                
                                if (sensorOldValue != sensorNewValue) ic.setInterrupt(1);


				// message(getProcessName() + ": SLL=" + hw.getStationLevelLow() + " SLH=" + hw.getStationLevelHigh());
				// message(getProcessName() + ": Output flow = " + hw.getOutputFlow() + " liters per second.");
			} // while true
		} // try
		catch (JSimSecurityException e1)
		{
			e1.printComment(System.err);
			e1.printStackTrace();
		} // catch
	} // life


} // class StationProcess

