package cz.zcu.fav.kiv.jsimcasestudies.watersystem.model;

import cz.zcu.fav.kiv.jsim.*;

/**
 *  Dynamic model of water consumption.
 *  This process is responsible for periodic adjustments of the output flow from the main tank caused by
 *  people's consumption. It only updates the Output Flow register and decrements the water volume
 *  in the main tank.
 *
 *  @author Cesar ALLAIN
 *  @author Jarda KACER
 *  @version 1.2 from 2004-03-24
 */

class WaterConsumptionProcess extends JSimProcess
{
	/**
	 *  Multiplier of the model's base period. The product of this multiplier and the model's base
	 *  period will be used as a parameter to the hold() method.
	 */
	private static final double BASE_PERIOD_MULTIPLIER = 1.0;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  The model's base period of updates.
	 */
	private final double basePeriod;

	/**
	 *  Information produced by hardware sensors.
	 */
	private StationHardwareData hw;

	/**
	 *  The station that this process manages.
	 */
	private Station station;

	/* ------------------------------------------------------------------------------------------------ */

	/**
	 *  Constructs a new process simulating changes of the output flow from the main tank.
	 *  These changes are caused by varying people's water consumption.
	 *
	 *  @param simulation		The overall simulation.
	 *  @param hw			Hardware data periodically changed by the tank's flow sensor.
	 *  @param station		The station to be managed.
	 *  @param basePeriod		The model's base period.
	 */
	protected WaterConsumptionProcess(JSimSimulation simulation, StationHardwareData hw, Station station, double basePeriod) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
	{
		super("Water Consumption - Output Flow Changes", simulation);
		this.hw = hw;
		this.station = station;
		this.basePeriod = basePeriod;
	} // constructor

	/**
	 *  Simulates output flow changes in the pipe leading from the main tank. Reads and sets
	 *  the Output Flow register in the station's hardware data.
	 */
	protected void life()
	{
		try
		{
			double currentOutputFlow;
			double oldOutputFlow;
			double holdTime = basePeriod * BASE_PERIOD_MULTIPLIER;

			while (true)
			{
				// message(getProcessName() + ": Holding for " + holdTime);
				hold(holdTime);
                                // hold(JSimSystem.negExp(1.0/holdTime)); // for testing

				oldOutputFlow = (double) hw.getOutputFlow();
				station.decreaseVolume(oldOutputFlow * holdTime);
				currentOutputFlow = oldOutputFlow + randomOutputFlowChange(holdTime);
				// message(getProcessName() + ": Volume in the main tank: " + station.getVolume());

				if (currentOutputFlow < 0.0)
					currentOutputFlow = 0.0;
				if (currentOutputFlow > Station.MAX_OUTPUT_FLOW)
					currentOutputFlow = Station.MAX_OUTPUT_FLOW;
				if (station.getVolume() <= 0.0)
					currentOutputFlow = 0.0;

				hw.setOutputFlow((int) Math.rint(currentOutputFlow));
				hw.setStationLevelLow(station.getVolume() >= Station.LOW_VOLUME_LIMIT);
				hw.setStationLevelHigh(station.getVolume() >= Station.HIGH_VOLUME_LIMIT);
				// message(getProcessName() + ": SLL=" + hw.getStationLevelLow() + " SLH=" + hw.getStationLevelHigh());
				// message(getProcessName() + ": Output flow = " + hw.getOutputFlow() + " liters per second.");
			} // while
		} // try
		catch (JSimSecurityException e1)
		{
			e1.printComment(System.err);
			e1.printStackTrace();
		} // catch
	} // life

	/**
	 *  Computes a change of the output flow. A random number with uniform
	 *  distribution limited by -MAX_OUTPUT_FLOW_CHANGE_PER_SECOND from the left and
	 *  +MAX_OUTPUT_FLOW_CHANGE_PER_SECOND from the right is multiplied by the time interval
	 *  between two consecutive updates of model values.
	 *
	 *  @param time The time period during which the change happened.
	 *
	 *  @return A change of the output flow which has a uniform distribution.
	 */
	private double randomOutputFlowChange(double time)
	{
		double change;

		change = JSimSystem.uniform(-Station.MAX_OUTPUT_FLOW_CHANGE_PER_SECOND, +Station.MAX_OUTPUT_FLOW_CHANGE_PER_SECOND);
		change *= time;

		return change;
	} // randomOutputFlowChange

} // class WaterConsumptionProcess

