/*
 *  A complex example on scheduling possibilities. Several processes are created and activated.
 *  Every time they get a possibility to run they choose either to continue living or to die.
 *
 *  A process can also passivate itself, relying on its "neighbour" process which can possibly wake it up.
 *  This demonstrates possibilities of the passivate() and activate() functions.
 *
 *  An "ugly" process can also completely remove its neighbour from schedule which causes the neighbour
 *  not to be run anymore (except of when being scheduled again by activate()).
 *  This demonstrates possibilities of the cancel() function.
 *
 *  Sleeping time has exponential distribution with parameter Lambda.
 *  This demonstrates possibilities of the negExp() and hold() functions.
 *
 */


import cz.zcu.fav.kiv.jsim.*;

/*
 *  A simple main-program example
 */
public class SchedulingPossibilitiesExample
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation = null;
  SPE_Process[]         processes;
  int                   i;

  try
  {
   /*
    *  First, we create a simulation and five processes.
    */
   simulation = new JSimSimulation("Scheduling Possibilities Example");
   /*
    *  If you want to know what happens inside J-Sim and your simulation,
    *  uncomment the following line -- implicite value is false.
    */
   // simulation.setTalking(true);


   /*
    *  Now we create five processes and activate them.
    */
   processes = new SPE_Process[5];

   /* The first process has no neigbour to wake up*/
   processes[0] = new SPE_Process("Process No. " + Integer.toString(1), simulation, null);
   processes[0].activate(0.123);
   /* And nobody can wake up the fifth process */
   for (i = 1; i < processes.length; i++)
   {
    processes[i] = new SPE_Process("Process No. " + Integer.toString(i+1), simulation, processes[i-1]);
    processes[i].activate(0.123 + i*0.456);
   }


   /*
    *  Now we run the simulation. Again, this way of running a simulation is a very dangerous one.
    *  The fact whether we have a chance to exit the cycle or not depends on processes themselves, not on us.
    *  However here, we can be pretty sure that it's not a neverending loop -- each process
    *  will die sometimes.
    */
   while (simulation.step() == true)
    ;
  } // try

  /*
   *  All three possible exceptions are caught here in one handler
   */
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch

  /*
   *  It is a good practice to put the shutdown in the finally block. Then you can be sure
   *  that the JVM will not "hang".
   */
  finally
  {
   /*
    *  Every simulation must be shut down at the end of the program, even if an error occurs.
    */
   simulation.shutdown();
  }
 } // main
} // SchedulingPossibilitiesExample


/*
 *  A process example. During every loop, it may decide to activate its neighbour, to cancel it, to passivate itself,
 *  to to die or simply to sleep for a while.
 */
class SPE_Process extends JSimProcess
{
 private static final double ourCommonLambda            = 0.5;  // Parameter of sleeping time's distribution
 private static final double probabilityToDie           = 0.10; // Probability that process will die
 private static final double probabilityToPassivate     = 0.15; // Probability that process will passivate itself
 private static final double probabilityToActivate      = 0.20; // Probability that process will wake up its neighbour
 private static final double probabilityToCancel        = 0.05; // Probability that process will cancel its neighbour's schedule

 private SPE_Process     myNeighbour;
 /*
  *  Note that you must ALWAYS write your own constructor -- JSimProcess' one has two parameters
  */
 public SPE_Process(String name, JSimSimulation sim, SPE_Process neighbour) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
 {
  super(name, sim);
  myNeighbour = neighbour;
 } // constructor

 /*
  *  This is the life of the process. We will cycle here until we decide to die.
  */
 protected void life()
 {
  try
  {
   boolean      iWantToLive = true;
   double       timeToSleep;

   message("Hello! My name is " + getName() + ". I have just been born!");

   while (iWantToLive)
   {
    /*
     *  Shall we passivete ourselves?
     */
    if (Math.random() < probabilityToPassivate)
    {
     message("It is " + myParent.getCurrentTime() + " and " + getName() + " is passivating itself.");
     message("Maybe its neighbour will wake it up (maybe not).");
     passivate();
    } // if

    /*
     *  Shall we wake up our neighbour?
     */
    if ((Math.random() < probabilityToActivate) && (myNeighbour != null))
    {
     message("It is " + myParent.getCurrentTime() + " and " + getName() + " is activating its neighbour " + myNeighbour.getName() + ".");
     if (myNeighbour.isIdle())
     {
      myNeighbour.activate(myParent.getCurrentTime() + 0.01);
      message("OK, " + myNeighbour.getName() + " has been activated.");
     } // My neighbour was idle
     else
      message(myNeighbour.getName() + " was not activated. It had been either scheduled or terminated already.");
    } // if

    /*
     *  Shall we cancel our neighbour's schedule?
     */
    if ((Math.random() < probabilityToCancel) && (myNeighbour != null))
    {
     message("It is " + myParent.getCurrentTime() + " and " + getName() + " is cancelling its neighbour's (" + myNeighbour.getName() + ") schedule.");
     message("But there's still a chance to wake it up again.");
     if (myNeighbour.getState() == JSimProcess.STATE_SCHEDULED)
     {
      myNeighbour.cancel();
      message("OK, " + myNeighbour.getName() + " has been cancelled.");
     } // My neighbour was scheduled
     else
      message(myNeighbour.getName() + " was not cancelled because it had not been scheduled before.");
    } // if

    /*
     *  Shall we die or shall we continue after a little coffee break?
     */
    if (Math.random() < probabilityToDie)
    {
     iWantToLive = false;
     message("Good-bye, my friends. Sincerely yours, " + getName());
    }
    else
    {
     timeToSleep = JSimSystem.negExp(ourCommonLambda);
     message("It is " + myParent.getCurrentTime() + " and " + getName() + " is going to sleep for " + timeToSleep + " seconds.");
     hold(timeToSleep);
    } // else
   } // for i
  } // try
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch
 } // life
} // class SPE_Process

