/*
 *  A simple example, demonstrating J-Sim's possibilities to catch exception thrown in the life() method
 *  of your process. An ArithmeticException (division by zero) is used here as example.
 *
 *  In such a case, J-Sim destroys your process but keeps other processes running. A warning is printed out
 *  if talking is set to true.
 */


import cz.zcu.fav.kiv.jsim.*;

/*
 *  A simple main-program example
 */
public class ErrorInLife
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation = null;
  BadDivisionProcess    process = null;

  try
  {
   /*
    *  First, we create a simulation and one process.
    */
   simulation = new JSimSimulation("Error In Life Simulation");
   /*
    *  To obtain a J-Sim error message, talking must be set to true.
    *  If not set, we will just get a stack trace, usually printed to error output, looking like the following one:

        Hello!
        java.lang.ArithmeticException: / by zero
                at BadDivisionProcess.life(Compiled Code)
                at cz.zcu.fav.kiv.jsim.JSimProcess.run(JSimProcess.java:154)
    */
   //simulation.setTalking(true);

   process = new BadDivisionProcess("Bad Guy", simulation);

   /*
    *  We must not forget to activate the process. A nice time, isn't it?
    */
   process.activate(1.2345);

   /*
    *  Now we run the simulation. We will do as many steps as it is needed by all
    *  processes present in the system. This is potentially dangerous if there is at least
    *  one process which never wants to finish.
    */
   while (simulation.step() == true)
    ;
  } // try

  /*
   *  All three possible exceptions are caught here in one handler
   */
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch

  /*
   *  It is a good practice to put the shutdown in the finally block. Then you can be sure
   *  that the JVM will not "hang".
   */
  finally
  {
   /*
    *  Every simulation must be shut down at the end of the program, even if an error occurs.
    */
   simulation.shutdown();
  }
 } // main
} // ErrorInLife


/*
 *  A simple process example. It divides number 100 by ten different numbers and crashes at 100/0.
 *  Note that we must use integer division. Dividing a float number by zero returns Infinity which is not
 *  an error.
 */
class BadDivisionProcess extends JSimProcess
{
 /*
  *  Note that you must ALWAYS write your own constructor -- JSimProcess' one has two parameters
  */
 public BadDivisionProcess(String name, JSimSimulation sim) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
 {
  super(name, sim);
 } // constructor

 /*
  *  This is the life of the process. We will cycle here ten times, dividing numbers.
  */
 protected void life()
 {
  try
  {
   int   i;

   message("Hello! I have just been born!");

   for (i = 10; i >= 0; i--)
   {
    message("Hello!");
    message("100 divided by " + i + " is " + 100/i);
    /*
     *  This hold() is not necessary for this demonstration.
     */
    hold(9.8765);
   } // for i

   message("You should never see this message.");
  } // try
  catch (JSimSecurityException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch
 } // life
} // class BadDivisionProcess

