/*
 *  A simple example, demonstrating that a process may passivate only itself, not another one.
 *
 */


import cz.zcu.fav.kiv.jsim.*;

/*
 *  A simple main-program example
 */
public class ForbiddenPassivate
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation  = null;
  BadPassivateProcess   badProcess  = null, goodProcess = null;
  int                   i;

  try
  {
   /*
    *  First, we create a simulation and two processes.
    */
   simulation = new JSimSimulation("Forbidden Passivate Simulation");
   /*
    *  It is not necessary to set talking to true
    */
   // simulation.setTalking(true);

   goodProcess = new BadPassivateProcess("Good Guy", simulation, null);
   badProcess  = new BadPassivateProcess("Bad Guy", simulation, goodProcess);

   /*
    *  We must not forget to activate both processes.
    */
   goodProcess.activate(1.0);
   badProcess.activate(2.0);

   /*
    *  Now we do just five steps. It is enough to show the feature.
    */
   for (i = 0; (i < 5) && (simulation.step() == true); i++)
    ;
  } // try

  /*
   *  All three possible exceptions are caught here in one handler
   */
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch

  /*
   *  It is a good practice to put the shutdown in the finally block. Then you can be sure
   *  that the JVM will not "hang".
   */
  finally
  {
   /*
    *  Every simulation must be shut down at the end of the program, even if an error occurs.
    */
   simulation.shutdown();
  }
 } // main
} // ForbiddenPassivate


/*
 *  A simple process example. It tries to passivate another process but is not successful.
 */
class BadPassivateProcess extends JSimProcess
{
 private BadPassivateProcess    procToPassivate;

 /*
  *  Note that you must ALWAYS write your own constructor -- JSimProcess' one has two parameters
  */
 public BadPassivateProcess(String name, JSimSimulation sim, BadPassivateProcess proc) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
 {
  super(name, sim);
  procToPassivate = proc;
 } // constructor

 /*
  *  This is the life of the process. We will try to passivate the other guy or print hello.
  */
 protected void life()
 {
  try
  {
   int   i;

   /* Are we good or bad? */
   if (procToPassivate != null) // Bad
   {
    message(getName() + ": Hello! I'm going to call " + procToPassivate.getName() + "'s passivate() method!");
    procToPassivate.passivate();
    message(getName() + ": It's done, bye.");
   }
   else // Good
    while (true)
    {
     message(getName() + ": Hello! I'm still here!");
     hold(0.9);
    } // while
  } // try
  catch (JSimSecurityException e)
  {
   error(getName() + ": I'm so sorry! I will never do this again.");
   e.printStackTrace();
   e.printComment(System.err);
  } // catch
 } // life
} // class BadPassivateProcess

