/*
 *  An example demonstrating that a process may not call its own cancel() method.
 *  Also, a call to cancel() of a process which is not scheduled (but passive or terminated) is not permitted
 *  and an exception is thrown out.
 *
 */


import cz.zcu.fav.kiv.jsim.*;

/*
 *  A simple main-program example
 */
public class ForbiddenCancel
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation  = null;
  CancelProcess         process1  = null, process2 = null;
  int                   i;

  try
  {
   /*
    *  First, we create a simulation and two processes.
    */
   simulation = new JSimSimulation("Allowed Cancel Simulation");
   /*
    *  It is not necessary to set talking to true
    */
   // simulation.setTalking(true);

   process1 = new CancelProcess("Guy No 1", simulation, null);
   process2 = new CancelProcess("Guy No 2", simulation, process1);

   /*
    *  We must not forget to activate both processes.
    */
   process1.activate(1.0);
   process2.activate(2.0);

   /*
    *  Now we do just seven steps. It is enough to show the feature.
    */
   for (i = 0; (i < 7) && (simulation.step() == true); i++)
    ;
  } // try

  /*
   *  All three possible exceptions are caught here in one handler
   */
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  } // catch

  /*
   *  It is a good practice to put the shutdown in the finally block. Then you can be sure
   *  that the JVM will not "hang".
   */
  finally
  {
   /*
    *  Every simulation must be shut down at the end of the program, even if an error occurs.
    */
   simulation.shutdown();
  }
 } // main
} // ForbiddenCancel


/*
 *  A simple process example. It tries to passivate another process but is not successful.
 */
class CancelProcess extends JSimProcess
{
 private CancelProcess    procToCancel;

 /*
  *  Note that you must ALWAYS write your own constructor -- JSimProcess' one has two parameters
  */
 public CancelProcess(String name, JSimSimulation sim, CancelProcess proc) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
 {
  super(name, sim);
  procToCancel = proc;
 } // constructor

 /*
  *  This is the life of the process. We will try to passivate the other guy or print hello.
  */
 protected void life()
 {
  try
  {
   int   i;

   /* Are we good or bad? */
   if (procToCancel != null) // Bad
   {
    message("It is " + myParent.getCurrentTime() + " o'clock.");
    message(getName() + ": Hello! I'm going to cancel " + procToCancel.getName() + "!");

    /*
     *  This would have been perfectly OK if the process to be cancelled had been scheduled.
     *  But it is not because it has crashed already at simulation time 1.0.
     */
    procToCancel.cancel();
    message(getName() + ": It's done. He will not wake up anymore. Bye.");
   }
   else // Good
   {
    message(getName() + ": Hello!");
    message("It is " + myParent.getCurrentTime() + " o'clock.");
    message("I'm going now to cancel myself -- not permitted.");
    cancel();
    message("Done OK!"); // This will never appear
   } // else
  } // try
  catch (JSimSecurityException e)
  {
   error(getName() + ": Oh, no! I will never do this again.");
   e.printStackTrace();
   e.printComment(System.err);
  } // catch
 } // life
} // class CancelProcess

