/*
 *  A simple example using GUI output. Five processes are created and activated.
 *  When step() or message() is called for the first time, simulation switches to GUI mode.
 *  Unlike in previous examples, batch GUI mode is used here, which means that the user cannot
 *  control the simulation. The simulation behavior is hard-coded in the source file.
 */

import cz.zcu.fav.kiv.jsim.*;

public class GUIGreetings
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation = null;
  FiniteProcess         processA = null, processB = null, processC = null, processD = null, processE = null;
  long                  stepsExecuted = 0;

  try
   {
   /*
    *  As usual, we must create a simulation first.
    */
   System.out.println("Starting up simulation...");
   simulation = new JSimSimulation("Hello GUI World Simulation", JSimSimulation.MODE_GUI_BATCH);

   /*
    *  Uncomment the following line, if you want to see all details.
    */
   //simulation.setTalking(true);

   /*
    *  Now creating and activating processes in exactly the same way as we do it using console output.
    */
   simulation.message("Creating five processes...");
   processA = new FiniteProcess("Hello!", simulation, 0.5);
   processB = new FiniteProcess("Hello!", simulation, 0.6);
   processC = new FiniteProcess("Salut!", simulation, 1.5);
   processD = new FiniteProcess("Hej!",   simulation, 1.0);
   processE = new FiniteProcess("Ahoj!",  simulation, 1.2);

   processA.activate(0.0);
   processB.activate(0.5);
   processC.activate(1.0);
   processD.activate(1.2);
   processE.activate(1.5);

   // this simulation will terminate after some step() execution
   while (simulation.step() == true)
   {
    //stepsExecuted++;
    simulation.message("Step being executed " + stepsExecuted++ + ".");
   }
   simulation.message("Simulation terminated.");
   simulation.message("Steps executed: " + stepsExecuted);

   /*
    * This suspends the program until you close the main window.
    * without this you will not see any results after simulation's termination.
    */  
   simulation.waitForWindowClosure();
  
   simulation.talk("Back on the console, good-bye.");
  }

  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  }

  /*
   *  As usual, don't forget to shutdown the simulation. Do it always in the finally block.
   */
 finally
  {
   simulation.shutdown();
  }
 } // main
} // class Terminative



/*
 *  An example of simple process. Writes "Hello" or its equivalent and then sleeps.
 *  Sometimes it decides to die.
 *  This process could be used in a text-mode simulation too, there is no difference.
 */

class FiniteProcess extends JSimProcess
{
 private String         helloWord;
 private double         lambda;

 public FiniteProcess(String hello, JSimSimulation sim, double l) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException
 {
  super("Process that says " + hello, sim);
  helloWord = hello;
  lambda = l;
 } // constructor

 /*
  *  Note that we don't care at all about the mode we run in. We don't even know about it!
  */
 protected void life()
 {
  double        timeToSleep;

  try
  {
   /* To die or to live, that's the question. */
   while (JSimSystem.draw(0.95))
   {
    timeToSleep = JSimSystem.negExp(lambda);
    message(helloWord + " I'm going to sleep for " + timeToSleep + " seconds.");
    hold(timeToSleep);
   } // while

   error(helloWord + " I'm dying! Good-bye.");
  }
  catch (JSimSecurityException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  }
 } // life
} // class TerminativeProcess