/*
 *  This example shows how to replace the default content of the info window.
 *  See MyGUIDetailedInfoWindow and QueueWithServer classes for details.
 *
 *  A simple queueing network example. There are two queues, two servers and two generators of transactions.
 *  When a transaction is created, it is inserted into a queue (the corresponding queue of the generator that
 *  created the transaction) and later taken out from the queue by its server. The server then decides
 *  whether to throw the transaction away (the transaction disappears from the system completely) or whether
 *  to insert it into the other queue.
 *
 *  If you want to know what happens in the system, uncomment all calls to message() in all files.
 *
 */
import cz.zcu.fav.kiv.jsim.*;

public class QueueingNetworksExample
{
 public static void main(String[] args)
 {
  JSimSimulation        simulation = null;
  QueueWithServer       queue1 = null, queue2 = null;
  Server                server1 = null, server2 = null;
  Generator             generator1 = null, generator2 = null;
  double                mi1 = 1.0, mi2 = 1.0;
  double                lambda1 = 0.4, lambda2 = 0.4;
  double                p1 = 0.5, p2 = 0.5;

  try
  {
   System.out.println("Initializing the simulation..");

   simulation = new JSimSimulation("Queueing Networks Simulation", JSimSimulation.MODE_GUI_BATCH, 10);
   // simulation.setTalking(true);
   queue1 = new QueueWithServer("Queue 1", simulation, null);
   queue2 = new QueueWithServer("Queue 2", simulation, null);

   server1 = new Server("Server 1", simulation, mi1, p1, queue1, queue2);
   server2 = new Server("Server 2", simulation, mi2, p2, queue2, queue1);

   generator1 = new Generator("Generator 1", simulation, lambda1, queue1);
   generator2 = new Generator("Generator 2", simulation, lambda2, queue2);

   /*
    *  We must set the servers now because they didn't exist at the time the queues were created.
    */
   queue1.setServer(server1);
   queue2.setServer(server2);

   simulation.message("Activating the generators...");

   generator1.activate(0.0);
   generator2.activate(0.0);

   simulation.message("Running the simulation, please wait.");

   while ((simulation.getCurrentTime() < 2000.0) && (simulation.step() == true))
    ;

   /*
    *  Now, some boring numbers.
    */
   simulation.message("Simulation interrupted at time " + simulation.getCurrentTime());
   simulation.message("Queues' statistics:");
   simulation.message("Queue 1: Lw = " + queue1.getLw() + ", Tw = " + queue1.getTw() + ", Tw all = " + queue1.getTwForAllLinks());
   simulation.message("Queue 2: Lw = " + queue2.getLw() + ", Tw = " + queue2.getTw() + ", Tw all = " + queue2.getTwForAllLinks());
   simulation.message("Servers' statistics:");
   simulation.message("Server 1: Counter = " + server1.getCounter() + ", Tq (for transactions thrown away by this server) = " + server1.getTransTq());
   simulation.message("Server 2: Counter = " + server2.getCounter() + ", Tq (for transactions thrown away by this server) = " + server2.getTransTq());
   simulation.message("Mean response time = " + ( (server1.getTransTq()+server2.getTransTq()) / (server1.getCounter() + server2.getCounter()) ));

   simulation.waitForWindowClosure();
  } // try
  catch (JSimException e)
  {
   e.printStackTrace();
   e.printComment(System.err);
  }

  /*
   *  We must shutdown the simulation.
   */
  finally
  {
   simulation.shutdown();
  }
 } // main
} // class QueueingNetworksExample

