/*
 *  A simple simulation of two processes running in parallel and sharing a piece of global data
 *  protected with a semaphore. We are mainly interested in measuring the performance decrease caused
 *  by the necessary delays the two processes spend in the P() operation (until the other process
 *  leaves the critical section).
 */

import cz.zcu.fav.kiv.jsim.*;
import cz.zcu.fav.kiv.jsim.ipc.*;

public class CriticalSectionSimulation
{
	public static void main(String[] args)
	{
		JSimSimulation simulation = null;
		double mu = 1.0, lambda = 0.5;
		int countOfCriticalSectionUsage;
		double time;

		ProcessWithSemaphoreAndData p1, p2;
		SharedData data;
		JSimSemaphore semaphore;

		try
		{
			System.out.println("Initializing the simulation...");

			simulation = new JSimSimulation("Critical Section Simulation");
			data = new SharedData();
			semaphore = new JSimSemaphore("Semaphore to protect shared data", simulation, 1);
			p1 = new ProcessWithSemaphoreAndData("Process 1", simulation, mu, lambda, semaphore, data);
			p2 = new ProcessWithSemaphoreAndData("Process 2", simulation, mu, lambda, semaphore, data);

			simulation.message("Activating the processes...");

			p1.activate(0.0);
			p2.activate(0.0);

			simulation.message("Running the simulation, please wait.");

			while ((simulation.getCurrentTime() < 30000.0) && (simulation.step() == true))
				;

			// Now, some boring numbers.
			countOfCriticalSectionUsage = data.getCountOfCriticalSectionUsage();
			time = simulation.getCurrentTime();

			simulation.message("Simulation interrupted at time " + time);
			simulation.message("");
			// There are 2 processes, so we must divide the count by 2.
			simulation.message("Mean number of worker loops: " + (countOfCriticalSectionUsage / 2));
			// They ran in parallel (with the exception of the critical section), so we must give each the whole time.
			simulation.message("Mean time of worker loop: " + (2 * time / countOfCriticalSectionUsage));
			// f = 1/T, see the last computation.
			simulation.message("Mean frequency of worker loops: " + (countOfCriticalSectionUsage / (2 * time)));
			// Performance decrease = f_no_conficts / f_conflicts = T_conflicts / T_no_conflicts.
			// T_conflicts ... se above.
			// T_no_conflicts = 1/lambda + 1/mu = sum of mean times of both actions.
			simulation.message("Performance decrease caused by the critical section: " + ((2 * time / countOfCriticalSectionUsage) / (1/lambda + 1/mu)));
		} // try
		catch (JSimException e)
		{
			e.printStackTrace();
			e.printComment(System.err);
		} // catch

		// We must shutdown the simulation.
		finally
		{
			simulation.shutdown();
		}
	} // main

} // class CriticalSectionSimulation

