/*
 *	Copyright (c) 2000 - 2004, University of West Bohemia,
 *	                           Faculty of Applied Sciences,
 *	                           Department of Computer Science and Engineering,
 *	                           DSS Group.
 *	                           Pilsen, Czech Republic.
 *
 *	Please read the licence agreement carefully before using any part of this software.
 */
package cz.zcu.fav.kiv.jsim;

import java.io.PrintStream;
import java.util.*;

/**
 *	The JSimCalendar class holds information about all scheduled events within a J-Sim simulation.
 *	It contains a list of JSimCalendarEvent elements, ordered by their simulation time.
 *	J-Sim simulations use this class to schedule processes and to get information about a process which
 *	should be run during the next simulation step. You should never need to create an instance of this class.
 *
 *	@author		<a href="http://home.zcu.cz/~jkacer">Jarda KAER</a>
 *
 *	@version	J-Sim version 0.3.0
 *
 *	@since		J-Sim version 0.0.1
 */
public class JSimCalendar
{
	/**
	 *	Constant signalling that there is no process scheduled.
	 */
	public static final JSimProcess NOBODY = null;

	/**
	 *	Constant signalling that there is no process scheduled.
	 */
	public static final double NEVER = -1.0;

//--------------------------------------------------------------------------------------------------

	/**
	 *	This list holds all events of the calendar. Events are always kept ordered, because
	 *	the list is sorted after every insertion.
	 */
	private LinkedList eventList;

//--------------------------------------------------------------------------------------------------

	/**
	 *	Creates a new empty calendar.
	 */
	public JSimCalendar()
	{
		eventList = new LinkedList();
	} // constructor

	/**
	 *	Adds a new event with the specified time and the specified process to the calendar.
	 *
	 *	@param absTime	The absolute simulation time of the event.
	 *	@param process	The process to be later activated.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the absolute simulation
	 *							time is negative or the process is not specified (null).
	 */
	public synchronized void addEntry(double absTime, JSimProcess process) throws JSimInvalidParametersException
	{
		JSimCalendarEvent newEntry;

		if (absTime < 0.0)
			throw new JSimInvalidParametersException("JSimCalendar.addEntry(): absTime");
		if (process == null)
			throw new JSimInvalidParametersException("JSimCalendar.addEntry(): process");

		newEntry = new JSimCalendarEvent(absTime, process);
		eventList.add(newEntry);
		Collections.sort(eventList);
	} // addEntry

	/**
	 *	Adds the specified event to the calendar. This is useful when you already have an event (from outside)
	 *	and want to add it to the calendar.
	 *
	 *	@param calendarEvent	The event to be added to the calendar.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the absolute simulation
	 *							time is negative or the process is not specified (null).
	 */
	public synchronized void addWholeEvent(JSimCalendarEvent calendarEvent) throws JSimInvalidParametersException
	{
		if (calendarEvent.getTime() < 0.0)
			throw new JSimInvalidParametersException("JSimCalendar.addWholeEvent(): calendarEvent.absTime");
		if (calendarEvent.getProcess() == null)
			throw new JSimInvalidParametersException("JSimCalendar.addWholeEvent(): calendarEvent.process");

		eventList.add(calendarEvent);
		Collections.sort(eventList);
	} // addWholeEvent

	/**
	 *	Deletes one or all events of a process from the calendar. The rest of J-Sim assures that
	 *	a process will not have more than one event in the calendar so the second atribute is
	 *	more or less useless.
	 *
	 *	@param process	The process whose event(s) are to be deleted.
	 *	@param all	A flag saying that not only one but all events of the given process should be deleted.
	 *
	 *	@return		The number of events deleted.
	 *
	 *	@exception JSimInvalidParametersException	This exception is thrown out if the process is
	 *							not specified (null).
	 */
	public synchronized int deleteEntries(JSimProcess process, boolean all) throws JSimInvalidParametersException
	{
		JSimCalendarEvent entry;
		ListIterator it;
		int noOfDeleted = 0;

		if (process == null)
			throw new JSimInvalidParametersException("JSimCalendar.deleteEntries(): process");

		it = eventList.listIterator(0);
		while (it.hasNext() && ((all == true) || (noOfDeleted < 1)))
		{
			entry = (JSimCalendarEvent) it.next();
			if (entry.getProcess() == process)
			{
				it.remove();
				noOfDeleted++;
			} // if
		} // while

		return noOfDeleted;
	} // deleteEntries

	/**
	 *	Returns the first scheduled process. Its calendar event is at the head of the event list because
	 *	it is always sorted.
	 *
	 *	@return	The first scheduled process or NOBODY if the calendar is empty.
	 */
	public synchronized JSimProcess getFirstProcess()
	{
		if (eventList.size() < 1)
			return NOBODY;
		else
			return ((JSimCalendarEvent) eventList.getFirst()).getProcess();
	} // getFirstProcess

	/**
	 *	Returns the time of the event being at the head of the calendar.
	 *
	 *	@return	The time of the event being at the head of the calendar or NEVER if the calendar is empty.
	 */
	public synchronized double getFirstProcessTime()
	{
		if (eventList.size() < 1)
			return NEVER;
		else
			return ((JSimCalendarEvent) eventList.getFirst()).getTime();
	} // getFirstProcessTime

	/**
	 *	Deletes the event at the head of the calendar and sets the head to the event which follows
	 *	the current head.
	 */
	public synchronized void jump()
	{
		if (eventList.size() > 0)
			eventList.removeFirst();
	} // jump

	/**
	 *	Says whether the calendar is empty. The calendar is empty if it contains no events.
	 *
	 *	@return	True if it is empty, false otherwise.
	 */
	public synchronized boolean isEmpty()
	{
		return (eventList.size() < 1);
	} // isEmpty

	/**
	 *	Returns an iterator over all events in the calendar. However, it will not allow you to modify
	 *	the calendar. The iterator is built up from a copy of the calendar so any changes to the iterator,
	 *	such as remove(), will left the calendar unchanged.
	 *
	 *	@return	A list iterator containing copies of all events of the calendar.
	 */
	public synchronized ListIterator getIterator()
	{
		LinkedList ll = null;
		int i;

		// LinkedLists provide only shallow copies, we must clone all events individually.
		ll = (LinkedList) eventList.clone();
		for (i = 0; i < eventList.size(); i++)
			ll.set(i, ((JSimCalendarEvent) eventList.get(i)).clone());

		return ll.listIterator();
	} // getIterator

	/**
	 *	Returns a string representation of the calendar.
	 *
	 *	@return	A string representation of the calendar.
	 */
	public String toString()
	{
		return eventList.toString();
	} // toString
} // class JSimCalendar

